;; bdf.el -- BDF editor on Mule

;; Copyright (C) 1992 Electrotechnical Laboratory, JAPAN.

;; Keywords: bdf, bitmap, mule

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;; Author: Ken'ichi HANDA <handa@etl.go.jp>

(require 'bitmap)

(defconst bdf-block-array
  '((?0 . "$B!&!&!&!&(B")
    (?1 . "$B!&!&!&"#(B")
    (?2 . "$B!&!&"#!&(B")
    (?3 . "$B!&!&"#"#(B")
    (?4 . "$B!&"#!&!&(B")
    (?5 . "$B!&"#!&"#(B")
    (?6 . "$B!&"#"#!&(B")
    (?7 . "$B!&"#"#"#(B")
    (?8 . "$B"#!&!&!&(B")
    (?9 . "$B"#!&!&"#(B")
    (?a . "$B"#!&"#!&(B")
    (?b . "$B"#!&"#"#(B")
    (?c . "$B"#"#!&!&(B")
    (?d . "$B"#"#!&"#(B")
    (?e . "$B"#"#"#!&(B")
    (?f . "$B"#"#"#"#(B")
    (?A . "$B"#!&"#!&(B")
    (?B . "$B"#!&"#"#(B")
    (?C . "$B"#"#!&!&(B")
    (?D . "$B"#"#!&"#(B")
    (?E . "$B"#"#"#!&(B")
    (?F . "$B"#"#"#"#(B")))

(defconst bdf-block-array1 bdf-block-array)
(defconst bdf-block-array2 (nthcdr 8 bdf-block-array))

(defun bdf-visible ()
  (interactive)
  (bdf-update-current-glyph)
  (bdf-to-bitmap bdf-current-glyph)
  (let (buffer-read-only)
    (setq attributed-region nil)
    (setq bdf-mark nil)
    (goto-char (point-min))
    (search-forward "BITMAP\n")
    (if (< (following-char) 127)
	(let ((p (point)))
	  (search-forward "ENDCHAR")
	  (beginning-of-line)
	  (delete-region p (point))
	  (bdf-encode)))
    (search-forward "ENDCHAR\n")
    (delete-region (point) (point-max))
    (setq i 0)
    (while (< i *hex-len*)
      (insert (aref *cmp* i) ?\n)
      (setq i (1+ i)))
    (goto-char (point-min))))

(defun bdf-update-current-glyph ()
  (goto-char (point-min))
  (search-forward "BITMAP\n")
  (if (< (following-char) 127)
      (let ((i 0))
	(while (not (looking-at "ENDCHAR"))
	  (setq p (point))
	  (end-of-line)
	  (aset bdf-current-glyph i (buffer-substring p (point)))
	  (forward-line 1)
	  (setq i (1+ i))))
    (let ((i 0) (j 0) ch blocks)
      (while (/= (setq ch (following-char)) ?E)
	(if (= ch ?\n)
	    (progn
	      (setq i (1+ i) j 0)
	      (forward-char 1))
	  (setq blocks (if (= ch ?$B!&(B) bdf-block-array1 bdf-block-array2))
	  (while (and blocks
		      (not (looking-at (cdr (car blocks)))))
	    (setq blocks (cdr blocks)))
	  (aset (aref bdf-current-glyph i) j
		(if blocks (car (car blocks)) ?0))
	  (forward-char 4)
	  (setq j (1+ j)))))))

(defun bdf-unvisible ()
  (interactive)
  (bdf-update-current-glyph)
  (let (buffer-read-only p i)
    (setq attributed-region nil)
    (setq bdf-mark nil)
    (goto-char (point-min))
    (search-forward "BITMAP" nil t)
    (forward-line 1)
    (setq p (point))
    (search-forward "ENDCHAR" nil t)
    (beginning-of-line)
    (delete-region p (point))
    (setq i 0)
    (while (< i (length bdf-current-glyph))
      (insert (aref bdf-current-glyph i) ?\n)
      (setq i (1+ i)))
    (search-forward "ENDCHAR\n" nil t)
    (delete-region (point) (point-max)))
  (goto-char (point-min)))

(defun bdf-encode ()
  (let (i j line)
    (setq i 0)
    (while (< i (length bdf-current-glyph))
      (setq line (aref bdf-current-glyph i))
      (setq j 0)
      (while (< j (length line))
	(insert (cdr (assq (aref line j) bdf-block-array)))
	(setq j (1+ j)))
      (insert ?\n)
      (setq i (1+ i)))))


(defconst bdf-mode-map nil)
(make-variable-buffer-local 'mouse-map)
(if bdf-mode-map
    nil
  (setq bdf-mode-map (make-keymap))
  (suppress-keymap bdf-mode-map)
  (define-key bdf-mode-map "n" 'bdf-next-char)
  (define-key bdf-mode-map "p" 'bdf-prev-char)
  (define-key bdf-mode-map "N" 'bdf-next-ku)
  (define-key bdf-mode-map "P" 'bdf-prev-ku)
  (define-key bdf-mode-map "j" 'bdf-jump-char)
  (define-key bdf-mode-map "v" 'bdf-visible)
  (define-key bdf-mode-map "V" 'bdf-unvisible)
  (define-key bdf-mode-map "u" 'bdf-undo)
  (define-key bdf-mode-map "c" 'bdf-copy-glyph)
  (define-key bdf-mode-map "y" 'bdf-yank-glyph)
  (define-key bdf-mode-map "Y" 'bdf-yank-glyph-or)
  (define-key bdf-mode-map "t" 'bdf-top)
  (define-key mouse-map x-button-left 'bdf-pixel-set)
  (define-key mouse-map x-button-middle 'bdf-pixel-toggle)
  (define-key mouse-map x-button-right 'bdf-pixel-reset)
  (define-key mouse-map x-button-c-left 'bdf-start-region)
  (define-key mouse-map x-button-c-middle 'bdf-paste)
  (define-key mouse-map x-button-c-right 'bdf-end-region)
  (define-key mouse-map x-button-left-up 'bdf-update-mark)
  (define-key mouse-map x-button-middle-up 'bdf-update-mark)
  (define-key mouse-map x-button-right-up 'bdf-update-mark)
  (define-key mouse-map x-button-c-left-up 'bdf-update-mark)
  (define-key mouse-map x-button-c-middle-up 'bdf-update-mark)
  (define-key mouse-map x-button-c-right-up 'bdf-update-mark)
  )

(defvar bdf-glyph-register nil)

(defun bdf-mode ()
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'bdf-mode)
  (setq mode-name "BDF")
  (setq buffer-read-only t)
  (make-local-variable 'bdf-current-glyph)
  (setq bdf-current-glyph nil)
  (make-local-variable 'bdf-mark)
  (setq bdf-mark nil)
  (setq buffer-auto-save-file-name nil)
  (if (boundp 'mode-line-modified)
      (setq mode-line-modified "--- ")
    (setq mode-line-format
	  (cons "--- " (cdr (defautl-value 'mode-line-format)))))
  (use-local-map bdf-mode-map)
  (bdf-top)
  (search-forward "FONTBOUNDINGBOX " nil t)
  (let ((height (read (current-buffer)))
	(width (* (/ (+ (read (current-buffer)) 7) 8) 2))
	(i 0))
    (setq bdf-current-glyph (make-vector height nil))
    (while (< i height)
      (aset bdf-current-glyph i (make-string width ?0))
      (setq i (1+ i))))
  (goto-char (point-min)))

(defun bdf-top ()
  (interactive)
  (widen)
  (goto-char (point-min))
  (bdf-narrow-to-current-char))

(defun bdf-current-char ()
  (goto-char (point-min))
  (search-forward "STARTCHAR " nil t)
  (let ((p (point)))
    (end-of-line)
    (buffer-substring p (point))))

(defun bdf-narrow-to-current-char ()
  (let (p)
    (if (or (looking-at "STARTCHAR")
	    (search-backward "STARTCHAR" nil 'mv))
	(progn
	  (setq p (point))
	  (forward-line 1)
	  (search-forward "STARTCHAR" nil t)
	  (beginning-of-line)
	  (narrow-to-region p (point))
	  (setq mode-line-process (concat " " (bdf-current-char))))
      (setq p (point))
      (search-forward "STARTCHAR" nil t)
      (beginning-of-line)
      (narrow-to-region p (point))
      (setq mode-line-process " HEADER"))
    (goto-char p)
    (setq buffer-undo-list nil)))

(defun bdf-next-char ()
  (interactive)
  (let ((p1 (point-min))
	(p2 (point-max)))
    (goto-char p2)
    (widen)
    (if (search-forward "STARTCHAR" nil t)
	(bdf-narrow-to-current-char)
      (narrow-to-region p1 p2)
      (message "Already at the last char!"))))

(defun bdf-prev-char ()
  (interactive)
  (let ((p1 (point-min))
	(p2 (point-max)))
    (goto-char p1)
    (if (= p1 1)
	(message "Already at the header!")
      (widen)
      (if (search-backward "STARTCHAR" nil 'mv)
	  (bdf-narrow-to-current-char)
	(bdf-top)))))

(defun bdf-next-ku ()
  (interactive)
  (widen)
  (if (looking-at "STARTCHAR")
      (forward-line 1))
  (re-search-forward "STARTCHAR ..21" nil t)
  (bdf-narrow-to-current-char))

(defun bdf-prev-ku ()
  (interactive)
  (goto-char (point-min))
  (widen)
  (re-search-backward "STARTCHAR ..21" nil 'mv)
  (bdf-narrow-to-current-char))

(defun bdf-jump-char (ch)
  (interactive "sSTARTCHAR ? ")
  (if (and (stringp ch)
	   (= (length ch) 4))
      (let ((target (format "STARTCHAR %s" ch))
	    (cur-ch (bdf-current-char)))
	(widen)
	(if (string< cur-ch ch)
	    (search-forward target nil t)
	  (search-backward target nil t))
	(bdf-narrow-to-current-char))
    (error "Invalid character %s" ch)))

(defun bdf-x-mouse-in-bdf (arg)
  (save-window-excursion
    (save-excursion
      (x-mouse-select arg)
      (eq major-mode 'bdf-mode))))

(defun bdf-pixel-set (arg)
  (bdf-pixel-on-off arg t))

(defun bdf-pixel-toggle (arg)
  (bdf-pixel-on-off arg 'toggle))

(defun bdf-pixel-reset (arg)
  (bdf-pixel-on-off arg nil))

(defun bdf-update-mark (arg)
  (interactive)
  (if (bdf-x-mouse-in-bdf arg)
      (let (buffer-read-only)
	(if bdf-mark
	    (progn
	      (goto-char bdf-mark)
	      (if (or (looking-at "$B!&(B") (looking-at "$B"#(B"))
		  (attribute-on-region 'inverse (point) (+ (point) 3)))))
	(goto-char (point-min)))))

(defun bdf-pixel-on-off (arg set)
  (if (bdf-x-mouse-in-bdf arg)
      (progn
	(x-mouse-set-point arg)
	(let (buffer-read-only
	      (ch (following-char)))
	  (cond ((and (not (eq set t)) (= ch ?$B"#(B))
		 (delete-char 1)
		 (insert "$B!&(B"))
		((and set (= ch ?$B!&(B))
		 (delete-char 1)
		 (insert "$B"#(B"))))
	(goto-char (point-min)))
    (x-mouse-set-point arg)))

(defun bdf-inverse-rectangle (from to)
  (save-excursion
    (let (temp)
      (if (> from to) (setq temp to to from from temp)))
    (operate-on-rectangle 'bdf-inverse-line from to nil)))

(defun bdf-inverse-line (start before after)
  (attribute-on-region 'inverse start (point)))

(defun bdf-start-region (arg)
  (interactive)
  (if (bdf-x-mouse-in-bdf arg)
      (progn
	(x-mouse-set-point arg)
	(setq attributed-region nil)
	(setq bdf-mark nil)
	(if (or (looking-at "$B!&(B") (looking-at "$B"#(B"))
	    (let ((p (point))
		  buffer-read-only)
	      (setq bdf-mark p)
	      (forward-char 1)
	      (attribute-on-region 'inverse p (point))))
	(goto-char (point-min)))))

(defun bdf-end-region (arg)
  (interactive)
  (if (bdf-x-mouse-in-bdf arg)
      (progn
	(x-mouse-set-point arg)
	(if (and bdf-mark
		 (or (looking-at "$B!&(B") (looking-at "$B"#(B")))
	    (let (buffer-read-only)
	      (forward-char 1)
	      (setq bdf-glyph-register (extract-rectangle bdf-mark (point)))
	      (setq attributed-region nil)
	      (bdf-inverse-rectangle bdf-mark (point))
	      (bdf-show-glyph-register bdf-glyph-register)))
	(goto-char (point-min)))))

(defun bdf-show-glyph-register (rect)
  (let ((buf (current-buffer)))
    (pop-to-buffer (get-buffer-create "*bdf-glyph-register*"))
    (erase-buffer)
    (insert-rectangle rect)
    (pop-to-buffer buf)))

(defun bdf-undo ()
  (interactive)
  (let (buffer-read-only)
    (undo)
    (goto-char (point-min))))

(defun bdf-paste (arg)
  (interactive)
  (if (and (bdf-x-mouse-in-bdf arg)
	   bdf-glyph-register)
      (let (buffer-read-only)
	(x-mouse-set-point arg)
	(if (or (looking-at "$B!&(B") (looking-at "$B"#(B"))
	    (let ((height (length bdf-glyph-register))
		  (width (length (car bdf-glyph-register)))
		  (p (point)))
	      (end-of-line)
	      (if (or (> width (- (point) p))
		      (save-excursion
			(search-forward "ENDCHAR")
			(beginning-of-line)
			(> height (count-lines p (point)))))
		  (error "Rectangle too large!")
		(setq attributed-region nil)
		(setq bdf-mark nil)
		(goto-char (+ p width))
		(next-line (1- height))
		(delete-rectangle p (point))
		(goto-char p)
		(insert-rectangle bdf-glyph-register))))
	(goto-char (point-min)))))

(provide 'bdf)
